#include <SoftWire.h>
#include <AsyncDelay.h>

#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#include <EEPROM.h>

// Profiles

#define NPROFILES 4
uint8_t profileNr;


// Include register data for profiles

#define code PROGMEM
#define Reg_Store Reg_Store0
#define Reg_Data Reg_Data0
#include "Si5338-RevB-Registers-0.h"
#define Reg_Store Reg_Store1
#define Reg_Data Reg_Data1
#include "Si5338-RevB-Registers-1.h"
#define Reg_Store Reg_Store2
#define Reg_Data Reg_Data2
#include "Si5338-RevB-Registers-2.h"
#define Reg_Store Reg_Store3
#define Reg_Data Reg_Data3
#include "Si5338-RevB-Registers-3.h"

#undef Reg_Store

const Reg_Data* Reg_Stores[NPROFILES] = {(Reg_Data*)Reg_Store0, (Reg_Data*)Reg_Store1, (Reg_Data*)Reg_Store2, (Reg_Data*)Reg_Store3};
float f_xtal[NPROFILES] = { 26.0, 24.0, 10.0, 8.0 };

// Soft I2C
#define SOFT_SDA 10
#define SOFT_SCL 11
#define  SI_ADDR      0x71
//#define  WRITE             0x00           // SMBus WRITE command
//#define  READ              0x01           // SMBus READ comman
#define LOS_MASK 0x04
#define LOCK_MASK 0x15

SoftWire sw(SOFT_SDA, SOFT_SCL);

// These buffers must be at least as large as the largest read or write you perform.
char swTxBuffer[4];
char swRxBuffer[4];


// Display

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 32 // OLED display height, in pixels

#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// ADC

#define AREF_TYPE INTERNAL
#define AREF_VOLTAGE 1.1

// Buttons

#define BTN_UP 12
#define BTN_DOWN 9

// Soft I2C helper functions

void SoftI2C_ByteWrite(uint8_t addr, uint8_t data)
{

  sw.beginTransmission(SI_ADDR);
  sw.write(addr); 
  sw.write(data); 
  sw.endTransmission();  

/*
  sw.startWrite(SI_ADDR);
  sw.llWrite(addr);  
  sw.llWrite(data);
  sw.stop();
*/    
}

uint8_t SoftI2C_ByteRead(uint8_t addr)
{
  
  sw.beginTransmission(SI_ADDR);
  sw.write(addr); 
  sw.endTransmission();  
  sw.requestFrom(SI_ADDR, 1);

  uint8_t data = sw.read();
  return data;


/*
  sw.start(SI_ADDR, SoftWire::writeMode);
  sw.llWrite(addr);  

//  sw.stop(); 
  sw.start(SI_ADDR, SoftWire::readMode);
  uint8_t data;
  sw.llRead(data);
  sw.stop(); 
  return data;
*/
  
}

/*
si47x_status SI4735::getInterruptStatus()
{
    si47x_status status;

    waitToSend();
    Wire.beginTransmission(deviceAddress);
    Wire.write(GET_INT_STATUS);
    Wire.endTransmission();

    Wire.requestFrom(deviceAddress, 1);
    status.raw = Wire.read();

    return status;
}
*/

uint32_t msn_p1, msn_p2, msn_p3;
uint32_t ms_p1[4], ms_p2[4], ms_p3[4];
uint8_t r_div[4];

void startCollectData()
{
  msn_p1 = 0, msn_p2 = 0, msn_p3 = 0;
  ms_p1[0] = 0; ms_p1[1] = 0; ms_p1[2] = 0; ms_p1[3] = 0; 
  ms_p2[0] = 0; ms_p2[1] = 0; ms_p2[2] = 0; ms_p2[3] = 0; 
  ms_p3[0] = 0; ms_p3[1] = 0; ms_p3[2] = 0; ms_p3[3] = 0; 
  r_div[0] = 0; r_div[1] = 0; r_div[2] = 0; r_div[3] = 0;
}

void collectData(uint16_t reg, uint8_t value)
{
  switch( reg )
  {
    // R0
    case 31:
        r_div[0] = 1 << ((value>>2) & 0b111);
        break;
    // R1
    case 32:
        r_div[1] = 1 << ((value>>2) & 0b111);
        break;
    // R2
    case 33:
        r_div[2] = 1 << ((value>>2) & 0b111);
        break;
    // R3
    case 34:
        r_div[3] = 1 << ((value>>2) & 0b111);
        break;

    // MS0

    case 53:
        ms_p1[0] = value;
        break;
    case 54:
        ms_p1[0] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 55:
        ms_p1[0] |= ((((uint32_t)value) & 0b11) << 16);
        ms_p2[0] = ((((uint32_t)value) & 0b11111100) >> 2);
        break;
    case 56:
        ms_p2[0] |= ((((uint32_t)value) & 0b11111111) << 6);
        break;
    case 57:
        ms_p2[0] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 58:
        ms_p2[0] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 59:
        ms_p3[0] = value;
        break;
    case 60:
        ms_p3[0] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 61:
        ms_p3[0] |= ((((uint32_t)value) & 0b11111111) << 16);
        break;
    case 62:
        ms_p3[0] |= ((((uint32_t)value) & 0b00111111) << 24);
        break;

    // MS1

    case 64:
        ms_p1[1] = value;
        break;
    case 65:
        ms_p1[1] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 66:
        ms_p1[1] |= ((((uint32_t)value) & 0b11) << 16);
        ms_p2[1] = ((((uint32_t)value) & 0b11111100) >> 2);
        break;
    case 67:
        ms_p2[1] |= ((((uint32_t)value) & 0b11111111) << 6);
        break;
    case 68:
        ms_p2[1] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 69:
        ms_p2[1] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 70:
        ms_p3[1] = value;
        break;
    case 71:
        ms_p3[1] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 72:
        ms_p3[1] |= ((((uint32_t)value) & 0b11111111) << 16);
        break;
    case 73:
        ms_p3[1] |= ((((uint32_t)value) & 0b00111111) << 24);
        break;

    // MS2

    case 75:
        ms_p1[2] = value;
        break;
    case 76:
        ms_p1[2] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 77:
        ms_p1[2] |= ((((uint32_t)value) & 0b11) << 16);
        ms_p2[2] = ((((uint32_t)value) & 0b11111100) >> 2);
        break;
    case 78:
        ms_p2[2] |= ((((uint32_t)value) & 0b11111111) << 6);
        break;
    case 79:
        ms_p2[2] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 80:
        ms_p2[2] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 81:
        ms_p3[2] = value;
        break;
    case 82:
        ms_p3[2] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 83:
        ms_p3[2] |= ((((uint32_t)value) & 0b11111111) << 16);
        break;
    case 84:
        ms_p3[2] |= ((((uint32_t)value) & 0b00111111) << 24);
        break;

    // MS3

    case 86:
        ms_p1[3] = value;
        break;
    case 87:
        ms_p1[3] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 88:
        ms_p1[3] |= ((((uint32_t)value) & 0b11) << 16);
        ms_p2[3] = ((((uint32_t)value) & 0b11111100) >> 2);
        break;
    case 89:
        ms_p2[3] |= ((((uint32_t)value) & 0b11111111) << 6);
        break;
    case 90:
        ms_p2[3] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 91:
        ms_p2[3] |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 92:
        ms_p3[3] = value;
        break;
    case 93:
        ms_p3[3] |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 94:
        ms_p3[3] |= ((((uint32_t)value) & 0b11111111) << 16);
        break;
    case 95:
        ms_p3[3] |= ((((uint32_t)value) & 0b00111111) << 24);
        break;



    // MSN
    case 97:
        msn_p1 = value;
        break;
    case 98:
        msn_p1 |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 99:
        msn_p1 |= ((((uint32_t)value) & 0b11) << 16);
        msn_p2 = ((((uint32_t)value) & 0b11111100) >> 2);
        break;
    case 100:
        msn_p2 |= ((((uint32_t)value) & 0b11111111) << 6);
        break;
    case 101:
        msn_p2 |= ((((uint32_t)value) & 0b11111111) << 14);
        break;
    case 102:
        msn_p2 |= ((((uint32_t)value) & 0b11111111) << 22);
        break;
    case 103:
        msn_p3 = value;
        break;
    case 104:
        msn_p3 |= ((((uint32_t)value) & 0b11111111) << 8);
        break;
    case 105:
        msn_p3 |= ((((uint32_t)value) & 0b11111111) << 16);
        break;
    case 106:
        msn_p3 |= ((((uint32_t)value) & 0b00111111) << 24);
        break;
       
  }
}

void init_si5338()
{
  startCollectData();
  
  SoftI2C_ByteWrite(230, 0x10);                   //OEB_ALL = 1
  SoftI2C_ByteWrite(241, 0xE5);                   //DIS_LOL = 1

  uint16_t counter;
  Reg_Data curr;
  uint8_t curr_chip_val, clear_curr_val, clear_new_val, combined, reg;

  Serial.print('1');

  byte* from_ptr = (byte*) Reg_Stores[profileNr];

  uint16_t page = 0;
  
  for(counter=0; counter<NUM_REGS_MAX; counter++)
  {
      //curr = Reg_Store[counter];

      byte* to_ptr = (byte*) (&curr);
      
      for (int i=0; i<sizeof(Reg_Data); i++)
      {
        *(to_ptr++) = pgm_read_byte(from_ptr++);
      }

/*      
       Serial.print(curr.Reg_Addr);
       Serial.print(' ');
       Serial.print(curr.Reg_Val);
       Serial.print(' ');
       Serial.println(curr.Reg_Mask);
*/  
      if(curr.Reg_Mask != 0x00) 
      {
        if(curr.Reg_Mask == 0xFF) 
        { 
            // do a write transaction only 
            // since the mask is all ones
            SoftI2C_ByteWrite(curr.Reg_Addr, curr.Reg_Val);

            if (curr.Reg_Addr == 255) // Switch page
              page = curr.Reg_Val;

            // Collect data

            collectData(curr.Reg_Addr | (page << 8), curr.Reg_Val);

        }
        else
        {
            curr_chip_val = SoftI2C_ByteRead(curr.Reg_Addr);
            clear_curr_val = curr_chip_val & ~curr.Reg_Mask;
            clear_new_val = curr.Reg_Val & curr.Reg_Mask;
            combined = clear_new_val | clear_curr_val;
            SoftI2C_ByteWrite(curr.Reg_Addr, combined);
        }
      }
  }

//  Serial.println(counter);
  Serial.print('2');

// check LOS alarm for the xtal input 
   // on IN1 and IN2 (and IN3 if necessary) - 
   // change this mask if using inputs on IN4, IN5, IN6
   _delay_ms(1);
   reg = SoftI2C_ByteRead(218) & LOS_MASK; 
   while(reg != 0){
      _delay_ms(1);
      reg = SoftI2C_ByteRead(218) & LOS_MASK;
   }
   SoftI2C_ByteWrite(49, SoftI2C_ByteRead(49) & 0x7F); //FCAL_OVRD_EN = 0
   SoftI2C_ByteWrite(246, 2);                      //soft reset 
   SoftI2C_ByteWrite(241, 0x65);                   //DIS_LOL = 0

  Serial.print('3');

   // wait for Si5338 to be ready after calibration (ie, soft reset)

   _delay_ms(10);

   //make sure the device locked by checking PLL_LOL and SYS_CAL
   reg = SoftI2C_ByteRead(218) & LOCK_MASK;
   while(reg != 0){
     reg = SoftI2C_ByteRead(218) & LOCK_MASK;
   }
  Serial.print('4');
  
   //copy FCAL values
   SoftI2C_ByteWrite(45, SoftI2C_ByteRead(235));
   SoftI2C_ByteWrite(46, SoftI2C_ByteRead(236));
   // clear bits 0 and 1 from 47 and 
   // combine with bits 0 and 1 from 237
   reg = (SoftI2C_ByteRead(47) & 0xFC) | (SoftI2C_ByteRead(237) & 3); 
   SoftI2C_ByteWrite(47, reg);
   SoftI2C_ByteWrite(49, SoftI2C_ByteRead(49) | 0x80); // FCAL_OVRD_EN = 1
   SoftI2C_ByteWrite(230, 0x00);                   // OEB_ALL = 0
   //------------------------------------------------------------  

  Serial.print('5');
}

void init_display()
{
  
  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;); // Don't proceed, loop forever
  }  
  // Clear the buffer
  display.clearDisplay();

  // Draw a single pixel in white
//  display.drawPixel(10, 10, SSD1306_WHITE);

  // Show the display buffer on the screen. You MUST call display() after
  // drawing commands to make them visible on screen!
//  display.display();
}

void init_adc()
{
  analogReference(AREF_TYPE);
}


void init_buttons()
{
  pinMode(BTN_UP, INPUT_PULLUP);
  pinMode(BTN_DOWN, INPUT_PULLUP);
}

uint8_t getBtnPressed()
{
  if (digitalRead(BTN_UP) == LOW)
    return BTN_UP;
  else 
  if (digitalRead(BTN_DOWN) == LOW)
    return BTN_DOWN;

  return 0;
}

void setDefaults()
{
  profileNr = 0;
  
  writeConfig();
}

void readConfig()
{
  uint8_t sig = EEPROM.read(0);
  if (sig != 0xAB)
  {
    setDefaults();
    return;  
  }

  profileNr = EEPROM.read(1);
}

void writeConfig()
{
  EEPROM.write(0, 0xAB);
  EEPROM.write(1, profileNr);
}

void setup() 
{
  Serial.begin(9600);
/*
 Serial.print("    SDA pin: ");
  Serial.println(int(SOFT_SDA));
  Serial.print("    SCL pin: ");
  Serial.println(int(SOFT_SCL));
*/

  readConfig();

  
  // Setup the SI5338

//  sw.setTimeout_ms(40);
//  sw.begin();


  sw.setTxBuffer(swTxBuffer, sizeof(swTxBuffer));
  sw.setRxBuffer(swRxBuffer, sizeof(swRxBuffer));
  
  sw.setDelay_us(5);
  sw.setTimeout(1000);
  sw.begin();
  
  _delay_ms(1);

  init_si5338();
/*  
    Serial.println(msn_p1);
    Serial.println(msn_p2);
    Serial.println(msn_p3);
    Serial.println(sizeof(float));
    Serial.println(sizeof(double));
    Serial.println(sizeof(long double));
*/
    
  _delay_ms(1);
  init_display();

  init_adc();

  init_buttons();

}

void loop() 
{
  display.clearDisplay();

// Display output frequencies

  float msN = (msn_p1+512+((float)msn_p2)/msn_p3)/128.0; // !!! Check the formula later
  float fVco = f_xtal[profileNr] * msN;
//  Serial.println(fVco);
  float ms0 = (ms_p1[0]+512+((float)ms_p2[0])/ms_p3[0])/128.0; // !!! Check the formula later
  float fOut0 = fVco / ms0 / r_div[0]; 
//  Serial.println(fOut0);
  float ms1 = (ms_p1[1]+512+((float)ms_p2[1])/ms_p3[1])/128.0; // !!! Check the formula later
  float fOut1 = fVco / ms1 / r_div[1]; 
//  Serial.println(fOut1);
  float ms2 = (ms_p1[2]+512+((float)ms_p2[2])/ms_p3[2])/128.0; // !!! Check the formula later
  float fOut2 = fVco / ms2 / r_div[2]; 
//  Serial.println(fOut2);
  float ms3 = (ms_p1[3]+512+((float)ms_p2[3])/ms_p3[3])/128.0; // !!! Check the formula later
  float fOut3 = fVco / ms3 / r_div[3]; 
//  Serial.println(fOut3);

  display.setTextSize(1); // Draw 1X-scale text
  display.setTextColor(SSD1306_WHITE);

  display.setCursor(0, 12);
  display.print(F("1) "));
  display.print(fOut0);
  display.print(F("M"));
  display.setCursor(0, 24);
  display.print(F("2) "));
  display.print(fOut1);
  display.print(F("M"));
  display.setCursor(64, 12);
  display.print(F("3) "));
  display.print(fOut2);
  display.print(F("M"));
  display.setCursor(64, 24);
  display.print(F("4) "));
  display.print(fOut3);
  display.print(F("M"));

// Osc frequency

  display.setCursor(0, 0);

  display.print(F("0) "));
  display.print(f_xtal[profileNr]);  
  display.print(F("M"));

  // Output tenperature

//  display.setTextSize(2); // Draw 2X-scale text
//  display.setTextColor(SSD1306_WHITE);

#define V_AVG 10 
  int sensorValue = 0;
  for (int i=0; i<V_AVG; i++)
    sensorValue += analogRead(A7);
  float voltage= sensorValue * (AREF_VOLTAGE / 1023.0)/V_AVG;

  int tc = voltage * 100;
  //display.println(F("scroll"));
/*  
  display.print(F("T= "));
  display.print(tc);
  display.print(char(32));
  display.print(char(247));
  display.print(F("C"));
*/
//  display.print(F(", T="));

  display.setCursor(64, 0);
  
  display.print(tc);
  display.print(char(247));
  display.print(F("C"));

  // Select profile

  if (getBtnPressed()==BTN_UP)
  {
    
    profileNr++;
    profileNr %= NPROFILES; 

    writeConfig(); // Save settings
    init_si5338(); // Re-init
    
    while (getBtnPressed()) {}
    delay(100);
    
    return; // Re-draw
  }

  // Output profile number

/*
  display.setCursor(64, 25);
  display.print(F("Profile= "));
  display.print(profileNr);  
*/
  display.print(F(" P"));
  display.print(profileNr);  



  display.display();      
  delay(1000);
  
  return;

  return;
for (int i=0; i<255; i++)
{
  
  Serial.print(i);
  Serial.print(' ');
  Serial.println(SoftI2C_ByteRead(i));
}
  
  //Serial.println(I2C_ByteRead(218));
  //_delay_ms(300);
  
  return;
zzz: goto zzz;
}
